/****************************************************************************/
/*                                                                          */
/* Program name: MQIsdp protocol sample application                         */
/*                                                                          */
/* Description: The application simulates a flow meter. The flow rate will  */
/*              randomly as the program exectutes and will be published at  */
/*              set intervals.                                              */
/*                                                                          */
/*  Statement:  Licensed Materials - Property of IBM                        */
/*                                                                          */
/*              MQSeries SupportPac IA93                                    */
/*              (C) Copyright IBM Corp. 2002                                */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Function:                                                                */
/*                                                                          */
/* This application runs for a specified time varying the flowrate at a     */
/* defined interval and publishing the data at a defined interval on topic  */
/* demo/c/flowrate.                                                         */
/* The democ application maybe run to update the bounds within which demo   */
/* is running. If the bounds are changed such that the flowrate is outside  */
/* of the new bounds then the flowrate will 'creep' towards the new bounds  */
/* until the flowrate is within the the bounds.                             */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Change history:                                                          */
/*                                                                          */
/* V1.0   19-02-2003  IRH  Initial release                                  */
/*                                                                          */
/*==========================================================================*/
/* Module Name: demo.c                                                      */
#include <stdio.h>
#include <string.h>
#ifdef WIN32
  #include <time.h>
  #include <windows.h>
#else  
  #include <stdlib.h>
  #include <sys/time.h>
  #include <unistd.h>
  #include <stdarg.h>
  #include <sys/types.h>
  #include <sys/ipc.h>
  #include <sys/sem.h>
  #include <pthread.h>
#endif
#include <MQIsdp.h>

#define NOT_FLOWING 0
#define FLOWING     1

#define MEAN_RATE 50
#define MAX_RATE  75
#define MIN_RATE  25

#define DEMO_TOPIC    "demo/c/flowrate"
#define DEMO_TOPIC_SZ 15
#define CONTROL_TOPIC "demo/c/control"
#define KEEP_ALIVE    15
#define CLIENT_ID     "MQISDP_C_DEMO"

#ifndef RAND_MAX
  #define RAND_MAX 32767
#endif

#define CaseRet( x ) \
  case x:            \
    ptr = #x;        \
    break;
  
typedef struct struc_mqisdp {
    MQISDPCH  hConn;
    long      keepAlive;
    char      infoString[MQISDP_INFO_STRING_LENGTH];
    long      infoCode;
    MQISDPTI  hMbs;
    MQISDPMH  lasthMsg;
    long      pubsSent;
    long      pubsRcvd;
}MQISDP_PARMS;

typedef struct struc_ppp {
    int    Duration;
    int    FlowState;
    time_t NextFlowStateChange;
    int    FlowingInterval;
    int    NotFlowingInterval;
    time_t NextFlowRateChange;
    int    FlowRateChangeInterval;
    int    FlowRate;
    int    MeanRate;
    int    MinRate;
    int    MaxRate;
    int    RateIncrement;
    time_t NextPublish;
    int    PublishInterval;
    int    PublishQoS;
    int    ReceiveInterval;
    int    NextReceive;
    char  *DemoTopic;
    char  *ControlTopic;
    char   IpAddress[MQISDP_INET_ADDR_LENGTH];
    long   IpPort;
    MQISDP_PARMS MqisdpParms;
} PROG_PARMS;
    
static int demoUsage( void );
static void demoLog( char *fmt, ... );
static void demoSleep( int mSecs );
static int demoInitialise( PROG_PARMS *ppp );
static int demoProcessCommandLine( PROG_PARMS *ppp, int argCount, char *argList[] );
static int demoChangeFlowState( PROG_PARMS *ppp );
static int demoChangeFlowRate( PROG_PARMS *ppp );
static int demoMQIsdpConnect( PROG_PARMS *ppp );
static int demoMQIsdpDisconnect( PROG_PARMS *ppp );
static int demoMQIsdpPublish( PROG_PARMS *ppp );
static int demoMQIsdpSubscribe( PROG_PARMS *ppp );
static int demoMQIsdpUnsubscribe( PROG_PARMS *ppp );
static int demoMQIsdpRcvPub( PROG_PARMS *ppp );
static int demoRand( int min, int max );
static int demoGetRCtoString( long rc, char *pBuffer, long bufSz );

int main( int argc, char **argv ) {
    PROG_PARMS pp;
    time_t     curTime;
    time_t     startTime;
    time_t     endTime;
    MQISDPTI  *pSendTaskParms;
    MQISDPTI  *pRcvTaskParms;
    int        statusCode = 0;
    char       rcBuf[MQISDP_RC_STRING_LENGTH];

    demoInitialise( &pp );

    setvbuf( stdout, NULL, _IONBF, 0 );
    setvbuf( stderr, NULL, _IONBF, 0 );

    if ( demoProcessCommandLine( &pp, argc, argv ) != 0 ) {
        exit(1);
    }

    /* Send and receive task parms must be on the heap */
    pSendTaskParms = (MQISDPTI*)malloc( sizeof(MQISDPTI) );
    pRcvTaskParms = (MQISDPTI*)malloc( sizeof(MQISDPTI) );
    if ( MQIsdp_StartTasks( &pp.MqisdpParms.hMbs, pSendTaskParms,
                            pRcvTaskParms, CLIENT_ID ) != 0 ) {
        exit(1);
    }

    pp.NextFlowStateChange = time(NULL) + pp.FlowingInterval;
    pp.NextFlowRateChange = time(NULL) + pp.FlowRateChangeInterval;
    pp.NextPublish = time(NULL) + pp.PublishInterval;
    pp.NextReceive = time(NULL) + pp.ReceiveInterval;
    
    /* Connect to the MQIsdp broker */
    if ( demoMQIsdpConnect( &pp ) != 0 ) {
        exit(2);
    }

    /* Subscribe to the control topic */
    demoMQIsdpSubscribe( &pp );

    time( &startTime );
    if ( pp.Duration != 0 ) {
        endTime = startTime + pp.Duration;
    } else {
        endTime = 0;
    }
    time( &curTime );
    while( (curTime < endTime) || (endTime == 0) ) {
        if ( curTime >= pp.NextFlowRateChange ) {
            demoChangeFlowRate( &pp );
        }

        if ( curTime >= pp.NextFlowStateChange ) {
            demoChangeFlowState( &pp );
        }

        if ( curTime >= pp.NextPublish ) {
            if ( demoMQIsdpPublish( &pp ) != 0 ) {
                /* End the test early when we detect an error */
                endTime = curTime;
            }
        }

        if ( curTime >= pp.NextReceive ) {
            demoMQIsdpRcvPub( &pp );
        }

        demoSleep( 500 );
    
        time( &curTime );
    }
    
    /* Receive remaining publications */
    demoMQIsdpRcvPub( &pp );
    
    /* Unsubscribe from the control topic */
    demoMQIsdpUnsubscribe( &pp );
    
    /* Wait until the unsubscribe has been delivered */
    statusCode = MQIsdp_getMsgStatus( pp.MqisdpParms.hConn, pp.MqisdpParms.lasthMsg );
    while ( (statusCode != MQISDP_DELIVERED) && (statusCode != MQISDP_DISCARDED) ) {
        printf( "Waiting for message handle %ld to be delivered....\n", pp.MqisdpParms.lasthMsg );
        demoSleep( 5000 );
        statusCode = MQIsdp_getMsgStatus( pp.MqisdpParms.hConn, pp.MqisdpParms.lasthMsg );
    }
    demoGetRCtoString( statusCode, rcBuf, MQISDP_RC_STRING_LENGTH );
    demoLog( "MQIsdp_getMsgStatus hMsg:%ld, status:%s\n", pp.MqisdpParms.lasthMsg, rcBuf );
    
    /* Receive remaining publications */
    demoMQIsdpRcvPub( &pp );
    
    demoMQIsdpDisconnect( &pp );

    MQIsdp_terminate();
    
    demoLog( "Publications Sent    :%ld\n", pp.MqisdpParms.pubsSent );
    demoLog( "Publications Received:%ld\n", pp.MqisdpParms.pubsRcvd );

    /* Wait for the tasks to terminate */
    demoSleep(7000);
    
    free( pSendTaskParms );
    free( pRcvTaskParms );
    free( pp.DemoTopic );
    free( pp.ControlTopic );
    
    return 0;
}

/* All intervals are in seconds */
static int demoInitialise( PROG_PARMS *ppp ) {

    ppp->FlowState = FLOWING;
    ppp->FlowingInterval = 120; 
    ppp->NotFlowingInterval = 60;

    ppp->FlowRate = MEAN_RATE;
    ppp->FlowRateChangeInterval = 5;
    
    ppp->PublishInterval = 5;
    ppp->DemoTopic = (char*)malloc( strlen(DEMO_TOPIC) + 1 );
    strcpy( ppp->DemoTopic, DEMO_TOPIC );
    ppp->ControlTopic = (char*)malloc( strlen(CONTROL_TOPIC) + 1 );
    strcpy( ppp->ControlTopic, CONTROL_TOPIC );
    ppp->PublishQoS = 0;
    
    ppp->Duration = 600;
    ppp->IpPort = 1883;

    ppp->MeanRate = MEAN_RATE;
    ppp->MaxRate = MAX_RATE;
    ppp->MinRate = MIN_RATE;
    /* Increment / decrement in units of 1/10th of the range */
    ppp->RateIncrement = (int)((ppp->MaxRate - ppp->MinRate) / 10);

    ppp->ReceiveInterval = 5;

    ppp->MqisdpParms.hConn = MQISDP_INV_CONN_HANDLE;
    ppp->MqisdpParms.keepAlive = KEEP_ALIVE;
    ppp->MqisdpParms.pubsSent = 0;
    ppp->MqisdpParms.pubsRcvd = 0;
    ppp->MqisdpParms.infoCode = 0;

    return 0;
}

static int demoProcessCommandLine( PROG_PARMS *ppp, int argCount, char *argList[] ) {
    int i;
    int rc = 2;
    /* Set to 0 if mandatory parm -a is specified */
    /* Set to 1 on error                          */

    for ( i=1; i < argCount; i++ ) {
        switch ( argList[i][1] ) {
        case 'd':
            /* Duration (seconds) */
            if ( strcmp( argList[i], "-d" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->Duration = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'f':
            /* Flowing interval (seconds) */
            if ( strcmp( argList[i], "-f" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->FlowingInterval = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'n':
            /* Not flowing interval (seconds) */
            if ( strcmp( argList[i], "-n" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->NotFlowingInterval = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'r':
            /* Rate change interval (seconds) */
            if ( strcmp( argList[i], "-r" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->FlowRateChangeInterval = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'p':
            /* Publish interval (seconds) */
            if ( strcmp( argList[i], "-p" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->PublishInterval = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'k':
            /* Keep alive interval (seconds) */
            if ( strcmp( argList[i], "-k" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->MqisdpParms.keepAlive = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'a':
            /* IP address */
            if ( strcmp( argList[i], "-a" ) == 0 ) {
                rc = 0;
                i++;
                if ( i < argCount ) {
                    strncpy( ppp->IpAddress, argList[i], MQISDP_INET_ADDR_LENGTH );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 's':
            /* Ip port number */
            if ( strcmp( argList[i], "-s" ) == 0 ) {
                i++;
                if ( i < argCount ) {
                    ppp->IpPort = atol( argList[i] );
                } else {
                    rc = 1;
                    demoUsage();
                }
            } else {
                demoUsage();
                rc = 1;
            }
            break;
        case 'h':
            demoUsage();
            rc = 1;
            break;
        default:
            printf( "Unrecognised flag:%s\n", argList[i] );
            demoUsage();
            rc = 1;
            break;
        }

        if ( rc == 1 ) {
            break;
        }
    }

    if ( rc == 2 ) {
        /* Mandatory parm missing */
        demoUsage();
    }
    
    return rc;
}

static int demoUsage( void ) {
    printf( "demo -a address <-s port> <-d duration> <-f flowing duration>\n" );
    printf( "     <-n not flowing duration> <-r interval> <-p publish interval>\n" );
    printf( "     <-k MQIsdp keep alive interval> <-h>\n" );
    printf( "\n" );
    printf( "  -d: Duration of demo (seconds). Default 600. 0 - run forever.\n" );
    printf( "  -k: MQIsdp keep alive interval (seconds). Default %ld.\n", KEEP_ALIVE );
    printf( "  -f: Duration of flowing state (seconds). Default 120.\n" );
    printf( "  -n: Duration of not flowing state (seconds). Default 60.\n" );
    printf( "  -r: Interval between flow rate changes (seconds). Default 5.\n" );
    printf( "  -p: Interval between flow rate publication (seconds). Default 5.\n" );
    printf( "  -a: TCP/IP address of MQIsdp broker e.g. 255.255.255.255\n" );
    printf( "  -s: TCP/IP port number of MQIsdp broker. Default 1883.\n" );
    printf( "  -h: Usage\n" );

    return 0;
}

static int demoChangeFlowState( PROG_PARMS *ppp ) {

    if ( ppp->FlowState == FLOWING ) {
        if ( ppp->NotFlowingInterval != 0 ) {
            ppp->FlowState = NOT_FLOWING;
            ppp->FlowRate = 0;
            ppp->NextFlowStateChange = time(NULL) + ppp->NotFlowingInterval;
        } else {
            ppp->NextFlowStateChange = time(NULL) + ppp->FlowingInterval;
        }
    } else {
        if ( ppp->FlowingInterval != 0 ) {
            ppp->FlowState = FLOWING;
            ppp->FlowRate = ppp->MeanRate;
            ppp->NextFlowStateChange = time(NULL) + ppp->FlowingInterval;
        } else {
            ppp->NextFlowStateChange = time(NULL) + ppp->NotFlowingInterval;
        }
    }

    return 0;
}

static int demoMQIsdpConnect( PROG_PARMS *ppp ) {
    CONN_PARMS  cp;
    int         rc = MQISDP_OK;
    long        connState = MQISDP_CONNECTING;
    char        rcBuf[MQISDP_RC_STRING_LENGTH];

    /* ####### CONNECT ####### */
    /* Sizeof connect structure:            */
    /* sizeof CONN_PARMS                    */
    /* No additional MQIsdp brokers         */
    /* No Will message                      */

    cp.version = 1;
    cp.strucLength = sizeof(CONN_PARMS);
    
    /* Add a client identifier */
    strncpy( cp.clientId, CLIENT_ID, MQISDP_CLIENT_ID_LENGTH );

    cp.retryCount = 10;
    cp.retryInterval = 10;
    cp.options = MQISDP_NONE; /* | MQISDP_CLEAN_START; */
    cp.keepAliveTime = (short)ppp->MqisdpParms.keepAlive;
    cp.numServers = 1;
    
    /* Mailbox handles */
    cp.apiMailbox = ppp->MqisdpParms.hMbs.apiMailbox;
    cp.sendMailbox = ppp->MqisdpParms.hMbs.sendMailbox;
    cp.sendMutex = ppp->MqisdpParms.hMbs.sendMutex;
    cp.receiveSemaphore = ppp->MqisdpParms.hMbs.receiveSemaphore;
    
    /* Port number */
    cp.mqisdpSvr[0].portNumber = ppp->IpPort;
    
    /* IP address */
    strncpy( cp.mqisdpSvr[0].ipAddress, ppp->IpAddress, MQISDP_INET_ADDR_LENGTH );

    rc = MQIsdp_connect( &(ppp->MqisdpParms.hConn), &cp );
    demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
    demoLog( "MQIsdp_connect rc:%s\n", rcBuf );
    
    while ( connState == MQISDP_CONNECTING ) {
        connState = MQIsdp_status(ppp->MqisdpParms.hConn, MQISDP_INFO_STRING_LENGTH,
                                  &(ppp->MqisdpParms.infoCode), ppp->MqisdpParms.infoString );
        demoGetRCtoString( connState, rcBuf, MQISDP_RC_STRING_LENGTH );
        demoLog( "MQIsdp_status state:%s - %s\n", rcBuf, ppp->MqisdpParms.infoString ); 
        demoSleep(1000);
    }

    /* If we haven't successfully connected to the MQIsdp broker then disconnect */
    if ( connState != MQISDP_CONNECTED ) {
        demoMQIsdpDisconnect( ppp );
        rc = 1;
    }

    return rc;
}

static int demoMQIsdpPublish( PROG_PARMS *ppp ) {
    MQISDPMH   hMsg;
    PUB_PARMS *pPp;
    int        rc = MQISDP_OK;
    char       rate[10];
    char       rcBuf[MQISDP_RC_STRING_LENGTH];

    pPp = (PUB_PARMS*)malloc( sizeof(PUB_PARMS) );

    if ( pPp != NULL ) {
        pPp->version = 1;
        pPp->strucLength = sizeof(PUB_PARMS);

        /* Set the quality of service for the publication */
        switch ( ppp->PublishQoS ) {
        case 0:
            pPp->options = MQISDP_QOS_0;
            break;
        case 1:
            pPp->options = MQISDP_QOS_1;
            break;
        case 2:
            pPp->options = MQISDP_QOS_2;
            break;
        }

        /* Cycle to the next QoS */
        (ppp->PublishQoS == 2) ? (ppp->PublishQoS = 0) : ppp->PublishQoS++;
    
        /* Add the topic name */
        pPp->topicLength = DEMO_TOPIC_SZ;
        pPp->topic = ppp->DemoTopic;

        /* Add the data */
        sprintf( rate, "%d", ppp->FlowRate );
        pPp->dataLength = strlen(rate);
        pPp->data = rate;

        /* Publish */
        rc = MQIsdp_publish( ppp->MqisdpParms.hConn, &hMsg, pPp );

        demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
        demoLog( "MQIsdp_publish rc:%s\n", rcBuf );
        
        if ( rc == MQISDP_OK ) {
            ppp->MqisdpParms.lasthMsg = hMsg;
            ppp->MqisdpParms.pubsSent++;
        } else if ( rc != MQISDP_Q_FULL ) {
            rc = 1;
        }
    
        free( pPp );
    }
    
    ppp->NextPublish = time(NULL) + ppp->PublishInterval;
    
    return rc;
}

static int demoMQIsdpDisconnect( PROG_PARMS *ppp ) {
    char rcBuf[MQISDP_RC_STRING_LENGTH];
    int  rc;

    rc = MQIsdp_disconnect( &(ppp->MqisdpParms.hConn) );

    demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
    demoLog( "MQIsdp_disconnect rc:%s\n", rcBuf );

    return rc;
}

static int demoMQIsdpSubscribe( PROG_PARMS *ppp ) {
    MQISDPMH   hSub;
    long       subLength;
    long       topicLength;
    long       QoS1 = MQISDP_QOS_1;
    long       QoS2 = MQISDP_QOS_2;
    SUB_PARMS *pSp;
    char      *pData;
    int        rc = MQISDP_OK;
    char       rcBuf[MQISDP_RC_STRING_LENGTH];

    /* SUBSCRIBE */
    /* SUB_PARMS structure */
    /* plus topic length field (long) */
    /* plus topic name field (string) */
    /* plus topic QoS field (long)    */
    topicLength = strlen(ppp->ControlTopic);
    subLength = sizeof(SUB_PARMS) + topicLength + (2*sizeof(long));
    
    pSp = (SUB_PARMS*)malloc( subLength );

    pSp->version = 1;
    pSp->strucLength = subLength;
    pData = (char*)pSp + sizeof(SUB_PARMS);

    /* Topic Length */
    memcpy( pData, &topicLength, sizeof(long) );
    pData += sizeof(long);

    /* Topic Name */
    memcpy( pData, ppp->ControlTopic, topicLength );
    pData += topicLength;
    
    /* Topic QoS */
    memcpy( pData, &QoS2, sizeof(long) );
    pData += sizeof(long);

    rc = MQIsdp_subscribe( ppp->MqisdpParms.hConn, &hSub, pSp );
    
    demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
    demoLog( "MQIsdp_subscribe rc:%s\n", rcBuf );
    
    free( pSp ); 

    return rc;
}

static int demoMQIsdpUnsubscribe( PROG_PARMS *ppp ) {
    MQISDPMH   hMsg;
    long       subLength;
    long       topicLength;
    UNSUB_PARMS *pUp;
    char      *pData;
    int        rc = MQISDP_OK;
    char       rcBuf[MQISDP_RC_STRING_LENGTH];

    /* UNSUBSCRIBE */
    /* UNSUB_PARMS structure */
    /* plus topic length field (long) */
    /* plus topic name field (string) */
    topicLength = strlen(ppp->ControlTopic);
    subLength = sizeof(UNSUB_PARMS) + topicLength + sizeof(long);
    
    pUp = (UNSUB_PARMS*)malloc( subLength );

    pUp->version = 1;
    pUp->strucLength = subLength;
    pData = (char*)pUp + sizeof(UNSUB_PARMS);

    /* Topic Length */
    memcpy( pData, &topicLength, sizeof(long) );
    pData += sizeof(long);

    /* Topic Name */
    memcpy( pData, ppp->ControlTopic, topicLength );
    pData += topicLength;
    
    rc = MQIsdp_unsubscribe( ppp->MqisdpParms.hConn, &hMsg, pUp );
    
    if ( rc == MQISDP_OK ) {
        ppp->MqisdpParms.lasthMsg = hMsg;
    }
    demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
    demoLog( "MQIsdp_unsubscribe rc:%s\n", rcBuf );
    
    free( pUp ); 

    return rc;
}

static int demoMQIsdpRcvPub( PROG_PARMS *ppp ) {
    long  bufferSz = 0;
    char *pBuffer = NULL;
    long  dataLength = 0;
    long  topicLength = 0;
    long  pubOptions = 0;
    int   rc;
    char *token;
    char  rcBuf[MQISDP_RC_STRING_LENGTH];

    /* ####### RECEIVE PUBLICATIONS */
    /* Loop while there are publications available */
    rc = MQISDP_PUBS_AVAILABLE;
    while ( (rc == MQISDP_PUBS_AVAILABLE) || (rc == MQISDP_DATA_TRUNCATED) ) {
        rc = MQIsdp_receivePub( ppp->MqisdpParms.hConn, 1000, &pubOptions, &topicLength, &dataLength, bufferSz, pBuffer );

        demoGetRCtoString( rc, rcBuf, MQISDP_RC_STRING_LENGTH );
        demoLog( "MQIsdp_receivePub rc:%s\n", rcBuf );
        
        switch ( rc ) {
        case MQISDP_DATA_TRUNCATED:
            if ( pBuffer == NULL ) {
                bufferSz = dataLength;
                pBuffer = (char*)malloc( bufferSz + 1 );
            } else {
                bufferSz = dataLength;
                pBuffer = (char*)realloc( pBuffer, bufferSz + 1 );
            }
            break;
        case MQISDP_PUBS_AVAILABLE:
        case MQISDP_OK:
            /* NULL terminate the data */
            pBuffer[dataLength] = '\0';
            ppp->MqisdpParms.pubsRcvd++;

            demoLog( "    Topic:%.*s\n",topicLength, pBuffer );
            demoLog( "    Data :%s\n", pBuffer + topicLength );
            
            token = strtok( pBuffer + topicLength, " " );
            while ( token != NULL ) {

                switch ( *token ) {
                case 'm':
                    token = strtok( NULL, " " );
                    if ( token != NULL ) {
                        ppp->MeanRate = atol(token);
                    }
                    break;
                case 'x':
                    token = strtok( NULL, " " );
                    if ( token != NULL ) {
                        ppp->MaxRate = atol(token);
                    }
                    break;
                case 'l':
                    token = strtok( NULL, " " );
                    if ( token != NULL ) {
                        ppp->MinRate = atol(token);
                    }
                    break;
                case 'r':
                    token = strtok( NULL, " " );
                    if ( token != NULL ) {
                        ppp->FlowRateChangeInterval = atol(token);
                    }
                    break;
                case 'p':
                    token = strtok( NULL, " " );
                    if ( token != NULL ) {
                        ppp->PublishInterval = atol(token);
                    }
                    break;
                default:
                    break;
                }

                token = strtok( NULL, " " );
            }

            break;
        default:
            break;
        }
    }

    /* Recalculate the increment with which the flow is varied */
    /* Increment / decrement in units of 1/10th of the range   */
    ppp->RateIncrement = (int)((ppp->MaxRate - ppp->MinRate) / 10);
    
    ppp->NextReceive = time(NULL) + ppp->ReceiveInterval;
    
    free( pBuffer );

    return 0;
}

static int demoChangeFlowRate( PROG_PARMS *ppp ) {
    int    rnum;

    if ( ppp->FlowState == FLOWING ) {
        rnum = demoRand( 1, 3 );

        switch ( rnum ) {
        case 3:
            /* Increase the flow rate */
            if ( ppp->FlowRate < ppp->MaxRate ) {
                ppp->FlowRate += ppp->RateIncrement;
                if ( ppp->FlowRate > ppp->MaxRate ) {
                    ppp->FlowRate = ppp->MaxRate;
                }
            }
            break;
        case 2:
            /* Keep the flow rate the same */
            break;
        case 1:
            /* Decrease the flow rate */
            if ( ppp->FlowRate > ppp->MinRate ) {
                ppp->FlowRate -= ppp->RateIncrement;
                if ( ppp->FlowRate < ppp->MinRate ) {
                    ppp->FlowRate = ppp->MinRate;
                }
            }
            break;
        }
    }

    ppp->NextFlowRateChange = time(NULL) + ppp->FlowRateChangeInterval;
    
    return 0;
}

static void demoLog( char *fmt, ... ) {
    char       buf[128];
    va_list    arg_ptr;

    va_start( arg_ptr, fmt );
    vsprintf ( buf, fmt, arg_ptr );
    va_end( arg_ptr );

    printf( buf );
}


/* Value RAND_MAX requires stdlib.h to be included */
static int demoRand( int min, int max ) {
    int    diff;
    int    randValue;
    double rand_double;
    double double_rand_max_plus_1 = (float) RAND_MAX + 1.0;

    randValue = rand();

    diff = max - min + 1;

    rand_double = ( double )randValue;

    randValue = (int)( ( rand_double / double_rand_max_plus_1 ) * diff ) + min;

    return( randValue );
}

static void demoSleep( int mSecs ) {
    #ifdef WIN32
      Sleep( mSecs );
    #endif
    
    #ifdef UNIX
      struct timeval tv;
      tv.tv_sec = (int)(mSecs / 1000);
      tv.tv_usec = (mSecs % 1000) * 1000;
      select( 0, NULL, NULL, NULL, &tv );
    #endif
}

static int demoGetRCtoString( long rc, char *pBuffer, long bufSz ) {
    
    char *ptr = NULL;

    switch ( rc ) {
       /* Return codes */
       CaseRet( MQISDP_OK )
       CaseRet( MQISDP_Q_FULL )
       CaseRet( MQISDP_FAILED )
       CaseRet( MQISDP_PUBS_AVAILABLE )
       CaseRet( MQISDP_NO_PUBS_AVAILABLE )
       CaseRet( MQISDP_CONN_HANDLE_ERROR )
       CaseRet( MQISDP_MSG_HANDLE_ERROR )
       CaseRet( MQISDP_NO_WILL_TOPIC )
       CaseRet( MQISDP_INVALID_STRUC_LENGTH )
       CaseRet( MQISDP_DATA_LENGTH_ERROR )
       CaseRet( MQISDP_DATA_TOO_BIG )
       CaseRet( MQISDP_ALREADY_CONNECTED )
       CaseRet( MQISDP_CONNECTION_BROKEN )
       CaseRet( MQISDP_DATA_TRUNCATED )
       CaseRet( MQISDP_OUT_OF_MEMORY )
        
       /* Message states */
       CaseRet( MQISDP_DELIVERED )
       CaseRet( MQISDP_RETRYING )
       CaseRet( MQISDP_IN_PROGRESS )
       CaseRet( MQISDP_DISCARDED )
       CaseRet( MQISDP_CONNECTING )
       CaseRet( MQISDP_CONNECTED )
       CaseRet( MQISDP_DISCONNECTED )
        
       /* Info codes */
       CaseRet( MQISDP_PROTOCOL_VERSION_ERROR )
       CaseRet( MQISDP_CLIENT_ID_ERROR )
       CaseRet( MQISDP_BROKER_UNAVAILABLE )
       CaseRet( MQISDP_SOCKET_CLOSED )
        default:
            break;
    }
    
    if ( pBuffer != NULL ) {
        if ( ptr != NULL ) {
            strncpy( pBuffer, ptr, bufSz );
        } else {
            sprintf( pBuffer, "%ld", rc );
        }
    }
    
    return rc;
}

